import asyncio
import json
from http import HTTPStatus

import pytest
import responses
from responses import matchers

from aioresponses import aioresponses
from pathlib import Path
from typing import Callable

import aiod
from aiod.calls.urls import server_url
from aiod.calls.utils import EndpointUndefinedError
from aiod.taxonomies import Term

resources_path = Path(__file__).parent / "resources"


asset_names = [
    "case_studies",
    "computational_assets",
    "contacts",
    "datasets",
    "educational_resources",
    "events",
    "experiments",
    "ml_models",
    "news",
    "organisations",
    "persons",
    "platforms",
    "projects",
    "publications",
    "services",
    "teams",
]

assets_with_search = [
    "datasets",
    "events",
    "experiments",
    "ml_models",
    "news",
    "organisations",
    "projects",
    "publications",
    "services",
]


@pytest.fixture(params=asset_names)
def asset_name(request):
    return request.param


@pytest.fixture(params=assets_with_search)
def asset_with_search(request):
    return request.param


def test_common_endpoints_are_created(asset_name: str):
    asset = getattr(aiod, asset_name)
    assert isinstance(getattr(asset, "get_list"), Callable)
    assert isinstance(getattr(asset, "counts"), Callable)
    assert isinstance(getattr(asset, "get_asset"), Callable)
    assert isinstance(getattr(asset, "get_asset_from_platform"), Callable)
    assert isinstance(getattr(asset, "get_content"), Callable)
    assert isinstance(getattr(asset, "get_list_async"), Callable)
    assert isinstance(getattr(asset, "get_assets_async"), Callable)


def test_search_endpoints_are_created(asset_with_search: str):
    asset = getattr(aiod, asset_with_search)
    assert isinstance(getattr(asset, "search"), Callable)


def test_endpoint_get_list(asset_name):
    with responses.RequestsMock() as mocked_requests:
        mocked_requests.add(
            responses.GET,
            f"{server_url()}{asset_name}?offset=0&limit=10",
            body=b'[{"resource_1": "info"},{"resource_2": "info"}]',
            status=200,
        )
        endpoint = getattr(aiod, asset_name)
        metadata_list = endpoint.get_list()

        assert len(metadata_list) == 2


def test_endpoint_counts(asset_name):
    with responses.RequestsMock() as mocked_requests:
        mocked_requests.add(
            responses.GET,
            f"{server_url()}counts/{asset_name}?detailed=false",
            body=b"2",
            status=200,
        )
        endpoint = getattr(aiod, asset_name)
        counts = endpoint.counts()

        assert counts == 2


def test_endpoint_get_asset(asset_name):
    with responses.RequestsMock() as mocked_requests:
        mocked_requests.add(
            responses.GET,
            f"{server_url()}{asset_name}/1",
            body=b'{"resource":"fake_details"}',
            status=200,
        )
        endpoint = getattr(aiod, asset_name)
        metadata = endpoint.get_asset(identifier=1, data_format="json")

        assert metadata == {"resource": "fake_details"}


@responses.activate
def test_endpoint_update_asset(asset_name, valid_refresh_token):
    responses.get(
        f"{server_url()}{asset_name}/1",
        body=b'{"name":"fake", "foo": "bar", "aiod_entry": {}}',
        status=200,
    )
    responses.put(
        f"{server_url()}{asset_name}/1",
        match=[matchers.json_params_matcher({"name":"fake", "foo": "baz"})],
        status=200,
    )
    endpoint = getattr(aiod, asset_name)
    endpoint.update(identifier=1, metadata=dict(foo="baz"))


def test_endpoint_get_list_from_platform(asset_name):
    platform_name = "zenodo"
    with responses.RequestsMock() as mocked_requests:
        mocked_requests.add(
            responses.GET,
            f"{server_url()}platforms/{platform_name}/{asset_name}?offset=0&limit=10",
            body=b'[{"resource_1": "info"},{"resource_2": "info"}]',
            status=200,
        )
        endpoint = getattr(aiod, asset_name)
        metadata_list = endpoint.get_list(platform=platform_name)

        assert len(metadata_list) == 2


def test_endpoint_get_asset_from_platform(asset_name):
    platform_name = "zenodo"
    platform_identifier = "zenodo.org:1000"
    with responses.RequestsMock() as mocked_requests:
        mocked_requests.add(
            responses.GET,
            f"{server_url()}platforms/{platform_name}/{asset_name}/{platform_identifier}",
            body=b'{"resource":"fake_details"}',
            status=200,
        )
        endpoint = getattr(aiod, asset_name)
        metadata = endpoint.get_asset_from_platform(
            platform=platform_name,
            platform_identifier=platform_identifier,
            data_format="json",
        )

        assert metadata == {"resource": "fake_details"}


def test_get_content(asset_name):
    with responses.RequestsMock() as mocked_requests:
        with open(resources_path / "content.csv", "r") as f:
            res_body = f.read()
        mocked_requests.add(
            responses.GET,
            f"{server_url()}{asset_name}/1/content",
            body=res_body,
            status=200,
        )
        endpoint = getattr(aiod, asset_name)
        content = endpoint.get_content(identifier=1)

        assert content == b"a,b,c\n1,2,3"


def test_get_content_with_distribution_idx(asset_name):
    with responses.RequestsMock() as mocked_requests:
        with open(resources_path / "content.csv", "r") as f:
            res_body = f.read()
        mocked_requests.add(
            responses.GET,
            f"{server_url()}{asset_name}/1/content/2",
            body=res_body,
            status=200,
        )
        endpoint = getattr(aiod, asset_name)
        content = endpoint.get_content(identifier=1, distribution_idx=2)

        assert content == b"a,b,c\n1,2,3"


def test_search(asset_with_search):
    search_query = "my query"
    search_field = "name"
    platforms = ["aiod", "openml"]
    get_all = True

    query = (
        f"?search_query={search_query.replace(' ', '+')}&platforms={platforms[0]}"
        f"&platforms={platforms[1]}&offset=0&limit=10&search_fields={search_field}&get_all=true"
    )
    with responses.RequestsMock() as mocked_requests:
        mocked_requests.add(
            responses.GET,
            f"{server_url()}search/{asset_with_search}{query}",
            body=b'{"total_hits": 2,"resources": [{"resource_1": "info"},{"resource_2": "info"}],"limit": 10,"offset": 0}',
            status=200,
        )
        endpoint = getattr(aiod, asset_with_search)
        metadata_list = endpoint.search(
            query=search_query,
            search_field=search_field,
            platforms=platforms,
            get_all=get_all,
            data_format="json",
        )

        assert len(metadata_list) == 2
        assert metadata_list == [{"resource_1": "info"}, {"resource_2": "info"}]


def test_endpoint_get_asset_async(asset_name):
    loop = asyncio.get_event_loop()
    with aioresponses() as mocked_responses:
        mocked_responses.get(
            f"{server_url()}{asset_name}/1",
            payload={"resource": "fake_details"},
            status=200,
        )
        mocked_responses.get(
            f"{server_url()}{asset_name}/3",
            payload={"resource": "fake_details_2"},
            status=200,
        )

        endpoint = getattr(aiod, asset_name)
        metadata = loop.run_until_complete(
            endpoint.get_assets_async(identifiers=[1, 3], data_format="json")
        )
        assert metadata == [
            {"resource": "fake_details"},
            {"resource": "fake_details_2"},
        ]


def test_endpoint_get_list_async(asset_name):
    loop = asyncio.get_event_loop()
    with aioresponses() as mocked_responses:
        mocked_responses.get(
            f"{server_url()}{asset_name}?offset=0&limit=2",
            payload=[{"resource_1": "info"}, {"resource_2": "info"}],
            status=200,
        )
        mocked_responses.get(
            f"{server_url()}{asset_name}?offset=2&limit=1",
            payload=[{"resource_3": "info"}],
            status=200,
        )

        endpoint = getattr(aiod, asset_name)
        metadata = loop.run_until_complete(
            endpoint.get_list_async(offset=0, limit=3, batch_size=2, data_format="json")
        )
        assert metadata == [
            {"resource_1": "info"},
            {"resource_2": "info"},
            {"resource_3": "info"},
        ]


@responses.activate
def test_register_asset(asset_name, valid_refresh_token):
    identifier = "data_123412341234123412341234"
    responses.post(
        f"http://not.set/not_set/{asset_name}",
        match=[matchers.header_matcher({"Authorization": "Bearer valid_access"})],
        json={"identifier": identifier},
    )
    module = getattr(aiod, asset_name)
    identifier_ = module.register(metadata=dict(name="Foo"))
    assert identifier_ == identifier


@responses.activate
def test_replace_asset(asset_name, valid_refresh_token):
    identifier = "data_123412341234123412341234"
    responses.put(
        f"http://not.set/not_set/{asset_name}/{identifier}",
        match=[matchers.header_matcher({"Authorization": "Bearer valid_access"})],
    )
    module = getattr(aiod, asset_name)
    res = module.replace(identifier=identifier, metadata=dict(description="Foo"))
    assert res.status_code == HTTPStatus.OK


@responses.activate
def test_update_asset_incorrect_identifier(asset_name, valid_refresh_token):
    identifier = "data_123412341234123412341234"
    responses.get(
        f"http://not.set/not_set/{asset_name}/{identifier}",
        match=[matchers.header_matcher({"Authorization": "Bearer valid_access"})],
        json={
            # Server has singular instead, e.g., 'Case_study' and not 'case_studies'.
            "detail": f"{asset_name} '{identifier}' not found in the database.",
            "reference": "8471b0ba3d74436ca645a6c49d0c7d65",
        },
        status=HTTPStatus.NOT_FOUND,
    )
    module = getattr(aiod, asset_name)
    with pytest.raises(KeyError) as e:
        module.update(identifier=identifier, metadata=dict(description="Foo"))
    msg = e.value.args[0]
    assert msg.startswith("No") and msg.endswith("found.")


@responses.activate
def test_delete_asset(asset_name, valid_refresh_token):
    identifier = "data_123412341234123412341234"
    responses.delete(
        f"http://not.set/not_set/{asset_name}/{identifier}",
        match=[matchers.header_matcher({"Authorization": "Bearer valid_access"})],
    )
    module = getattr(aiod, asset_name)
    res = module.delete(identifier=identifier)
    assert res.status_code == HTTPStatus.OK


@responses.activate
def test_delete_asset_incorrect_identifier(asset_name, valid_refresh_token):
    identifier = "data_123412341234123412341234"
    responses.delete(
        f"http://not.set/not_set/{asset_name}/{identifier}",
        match=[matchers.header_matcher({"Authorization": "Bearer valid_access"})],
        json={
            # Server has singular instead, e.g., 'Case_study' and not 'case_studies'.
            "detail": f"{asset_name} '{identifier}' not found in the database.",
            "reference": "8471b0ba3d74436ca645a6c49d0c7d65",
        },
        status=HTTPStatus.NOT_FOUND,
    )
    module = getattr(aiod, asset_name)
    with pytest.raises(KeyError) as e:
        module.delete(identifier=identifier)
    msg = e.value.args[0]
    assert msg.startswith("No") and msg.endswith("found.")


@responses.activate
def test_industrial_sector_taxonomy():
    taxonomy = json.loads((resources_path / "industrial_sectors.json").read_text())
    responses.get(f"http://not.set/not_set/industrial_sectors", json=taxonomy)

    industrial_sectors = aiod.taxonomies.industrial_sectors()
    car_industry = Term(
        term="Car Industry", taxonomy="industrial sectors", definition="", subterms=[]
    )
    assert industrial_sectors[0].term == "Manufacturing"
    assert car_industry in industrial_sectors[0].subterms
    assert len(industrial_sectors) == 21


@responses.activate
def test_taxonomy_not_found():
    responses.get(
        f"http://not.set/not_set/industrial_sectors",
        json={"detail": "Not Found"},
        status=HTTPStatus.NOT_FOUND,
    )
    with pytest.raises(EndpointUndefinedError):
        # Accessing through __wrapped__ to ensure the cache isn't hit.
        aiod.taxonomies.industrial_sectors.__wrapped__()



import json
import re
from http import HTTPStatus

import pytest
import responses

from aiod import config
import aiod.authentication.authentication as authentication


@pytest.fixture(autouse=True)
def setup_test_configuration(request):
    if "server" in request.keywords:
        return
    # User defaults are autoloaded, so we override them.
    config.api_server = "http://not.set/"
    config.version = "not_set"
    config.auth_server = "http://not.set/"
    config.realm = "not_set"
    config.client_id = "not_set"


@pytest.fixture(autouse=True)
def mocked_keycloak_server():
    url = "http://not.set/realms/not_set/protocol/openid-connect/token"

    def validate_refresh_token(request):
        match = re.match(
            r"client_id=(\w+)&grant_type=(\w+)&refresh_token=(\w+)", request.body
        )
        assert match, "Request type not yet added to mock."
        client, grant_type, token = match.groups()
        match (client, grant_type, token):
            case _, "refresh_token", "invalid":
                error = {
                    "error": "invalid_grant",
                    "error_description": "Invalid refresh token",
                }
                return HTTPStatus.BAD_REQUEST, {}, json.dumps(error)
            case _, "refresh_token", "expired":
                error = {
                    "error": "invalid_grant",
                    "error_description": "Offline user session not found",
                }
                return HTTPStatus.BAD_REQUEST, {}, json.dumps(error)
            case _, "refresh_token", "valid":
                body = {
                    "access_token": "valid_access",
                    "refresh_token": "valid",
                    "expires_in": 30,
                }
                return HTTPStatus.OK, {}, json.dumps(body)
            case unknown:
                raise RuntimeError(f"Unknown request payload: {unknown}")

    responses.add_callback(
        responses.POST,
        url,
        callback=validate_refresh_token,
    )
    yield


@pytest.fixture(autouse=True)
def no_token():
    # User token is autoloaded, so we override it.
    authentication._token = None


@pytest.fixture
def expired_refresh_token():
    authentication._token = authentication.Token(
        refresh_token="expired",
    )


@pytest.fixture
def valid_refresh_token():
    authentication._token = authentication.Token(
        refresh_token="valid",
    )


@pytest.fixture
def invalid_refresh_token():
    authentication._token = authentication.Token(
        refresh_token="invalid",
    )



from http import HTTPStatus

import aiohttp
import asyncio
import requests

import pandas as pd
from typing import Literal
from functools import partial

from aiod.authentication.authentication import get_token, _get_auth_headers
from aiod.configuration import config
from aiod.calls.urls import (
    url_to_get_asset,
    url_to_get_content,
    url_to_get_list,
    url_to_get_list_from_platform,
    url_to_get_asset_from_platform,
    url_to_resource_counts,
    url_to_search,
    server_url,
)
from aiod.calls.utils import format_response, wrap_calls, ServerError


def get_any_asset(
    identifier: str,
    data_format: Literal["pandas", "json"] = "pandas",
) -> dict | pd.Series:
    """Get any asset on AI-on-Demand by identifier.

    Parameters
    ----------
    identifier:
        The identifier of the asset, e.g., `'data_...'`.
    data_format:
        The format of the value to be returned.

    Returns
    -------
    :
        All metadata of the asset.

    Raises
    ------
    KeyError
        If the asset cannot be found.
    """
    res = requests.get(
        server_url() + f"assets/{identifier}",
        headers=_get_auth_headers(required=False),
        timeout=config.request_timeout_seconds,
    )

    if res.status_code == HTTPStatus.NOT_FOUND:
        raise KeyError(f"Asset with identifier {identifier!r} not found.")
    if res.status_code != HTTPStatus.OK:
        raise ServerError(res)
    return format_response(res.json(), data_format)


def get_list(
    *,
    asset_type: str,
    platform: str | None = None,
    offset: int = 0,
    limit: int = 10,
    version: str | None = None,
    data_format: Literal["pandas", "json"] = "pandas",
) -> pd.DataFrame | list[dict]:
    """Retrieve a list of ASSET_TYPE from the catalogue.

    All parameters must be specified by name.

    Parameters
    ----------
    platform
        Return metadata of ASSET_TYPE assets of this platform (default is None).
    offset
        The offset for pagination (default is 0).
    limit
        The maximum number of items to retrieve (default is 10).
    version
        The version of the endpoint (default is None).
    data_format
        The desired format for the response (default is "pandas").
        For "json" formats, the returned type is a json decoded type, i.e. in this case a list of dicts.

    Returns
    -------
    :
        The retrieved metadata in the specified format.
    """
    url = (
        url_to_get_list_from_platform(asset_type, platform, offset, limit, version)
        if platform is not None
        else url_to_get_list(asset_type, offset, limit, version)
    )
    res = requests.get(
        url,
        timeout=config.request_timeout_seconds,
    )
    resources = format_response(res.json(), data_format)
    return resources


def delete_asset(
    *,
    asset_type: str,
    identifier: str,
    version: str | None = None,
) -> requests.Response:
    """Delete ASSET_TYPE from the catalogue.

    All parameters must be specified by name.

    Parameters
    ----------
    version
        The version of the endpoint (default is None).

    Returns
    -------
    :
        The server response.
    """
    url = url_to_get_asset(asset_type, identifier, version)
    res = requests.delete(
        url,
        headers=get_token().headers,
        timeout=config.request_timeout_seconds,
    )
    if res.status_code == HTTPStatus.NOT_FOUND and "not found" in res.json().get(
        "detail"
    ):
        raise KeyError(f"No {asset_type} with identifier {identifier!r} found.")
    return res


def put_asset(
    *,
    asset_type: str,
    identifier: str,
    metadata: dict,
    version: str | None = None,
) -> requests.Response:
    """Replace a ASSET_TYPE in catalogue.

    All parameters must be specified by name.

    Notes
    -----
    Any attribute not specified in `metadata` will be replaced with the default value!
    If you wish to only modify some attributes and keep the values of others, make sure
    to provide _all_ asset metadata.

    Parameters
    ----------
    identifier
        The identifier of the asset whose metadata to replace.
    metadata
        A dictionary with for each attribute a value.
    version
        If provided, use this version of the REST API instead of `config.version`.

    Returns
    -------
    :
        The server response.

    Raises
    ------
        KeyError if the identifier is not known by the server.
    """
    url = url_to_get_asset(asset_type, identifier, version)
    res = requests.put(
        url,
        headers=get_token().headers,
        json=metadata,
        timeout=config.request_timeout_seconds,
    )
    if res.status_code == HTTPStatus.NOT_FOUND and "not found" in res.json().get(
        "detail"
    ):
        raise KeyError(f"No {asset_type} with identifier {identifier!r} found.")
    return res


def patch_asset(
    *,
    asset_type: str,
    identifier: str,
    metadata: dict,
    version: str | None = None,
) -> requests.Response:
    """Update an ASSET_TYPE in catalogue.

    All parameters must be specified by name.

    Notes
    -----
    This is a best-effort implementation, but is not yet officially supported by the server.

    Parameters
    ----------
    identifier
        The identifier of the asset whose metadata to replace.
    metadata
        A dictionary with for each attribute a value.
    version
        If provided, use this version of the REST API instead of `config.version`.

    Returns
    -------
    :
        The server response.

    Raises
    ------
        KeyError if the identifier is not known by the server.
    """
    url = url_to_get_asset(asset_type, identifier, version)

    asset = get_asset(
        identifier, asset_type=asset_type, version=version, data_format="json"
    )
    del asset["aiod_entry"]
    for attribute, value in metadata.items():
        asset[attribute] = value

    res = requests.put(
        url,
        headers=get_token().headers,
        json=asset,
        timeout=config.request_timeout_seconds,
    )
    if res.status_code == HTTPStatus.NOT_FOUND and "not found" in res.json().get(
        "detail"
    ):
        raise KeyError(f"No {asset_type} with identifier {identifier!r} found.")
    return res


def post_asset(
    *,
    asset_type: str,
    metadata: dict,
    version: str | None = None,
) -> str | requests.Response:
    """Register ASSET_TYPE in catalogue.

    All parameters must be specified by name.

    Parameters
    ----------
    metadata
        A dictionary with for each attribute a value.
    version
        If provided, use this version of the REST API instead of `config.version`.

    Returns
    -------
    identifier: str
        if the asset is registered successfully
    error response: requests.Response
        error response, if it failed to register successfully
    """
    url = f"{server_url(version)}{asset_type}"
    res = requests.post(
        url,
        headers=get_token().headers,
        json=metadata,
        timeout=config.request_timeout_seconds,
    )
    if res.status_code == HTTPStatus.OK:
        return res.json()["identifier"]
    return res


def counts(
    *, asset_type: str, version: str | None = None, per_platform: bool = False
) -> int | dict[str, int]:
    """Retrieve the number of ASSET_TYPE assets in the metadata catalogue.

    All parameters must be specified by name.

    Parameters
    ----------
    version
        The version of the endpoint (default is None).
    per_platform
        Whether to list counts per platform (default is False).

    Returns
    -------
    int | dict[str, int]
        The number ASSET_TYPE assets in the metadata catalogue.
        If the parameter per_platform is True,
        it returns a dictionary with platform names as keys
        and the number of ASSET_TYPE assets from that platform as values.
    """
    url = url_to_resource_counts(version, per_platform, asset_type)
    res = requests.get(url, timeout=config.request_timeout_seconds)
    return res.json()


def get_asset(
    identifier: str,
    *,
    asset_type: str,
    version: str | None = None,
    data_format: Literal["pandas", "json"] = "pandas",
) -> pd.Series | dict:
    """Retrieve metadata for a specific ASSET_TYPE.

    All parameters except `identifier` must be specified by name.

    Parameters
    ----------
    identifier
        The identifier of the ASSET_TYPE to retrieve.
    version
        The version of the endpoint (default is None).
    data_format
        The desired format for the response (default is "pandas").
        For "json" formats, the returned type is a json decoded type, in this case a dict.

    Returns
    -------
    :
        The retrieved metadata for the specified ASSET_TYPE.

    Raises
    ------
    KeyError
        If the asset cannot be found.
    """
    url = url_to_get_asset(asset_type, identifier, version)
    res = requests.get(
        url,
        headers=_get_auth_headers(required=False),
        timeout=config.request_timeout_seconds,
    )
    if res.status_code == HTTPStatus.NOT_FOUND and "not found" in res.json().get(
        "detail"
    ):
        raise KeyError(f"No {asset_type} with identifier {identifier!r} found.")
    resources = format_response(res.json(), data_format)
    return resources


def get_asset_from_platform(
    *,
    platform: str,
    platform_identifier: str,
    asset_type: str,
    version: str | None = None,
    data_format: Literal["pandas", "json"] = "pandas",
) -> pd.Series | dict:
    """Retrieve metadata for a specific ASSET_TYPE identified by the external platform identifier.

    All parameters must be specified by name.

    Parameters
    ----------
    platform
        The platform where the ASSET_TYPE asset is retrieved from.
    platform_identifier
        The identifier under which the ASSET_TYPE is known by the platform.
    version
        The version of the endpoint (default is None).
    data_format
        The desired format for the response (default is "pandas").
        For "json" formats, the returned type is a json decoded type, in this case a dict.

    Returns
    -------
    :
        The retrieved metadata for the specified ASSET_TYPE.
    """
    url = url_to_get_asset_from_platform(
        asset_type, platform, platform_identifier, version
    )
    res = requests.get(url, timeout=config.request_timeout_seconds)
    if res.status_code == HTTPStatus.NOT_FOUND and "not found" in res.json().get(
        "detail"
    ):
        raise KeyError(
            f"No {asset_type} with of {platform!r} with identifier {platform_identifier!r} found."
        )
    resources = format_response(res.json(), data_format)
    return resources


def get_content(
    *,
    identifier: str,
    asset_type: str,
    distribution_idx: int = 0,
    version: str | None = None,
) -> bytes:
    """Retrieve the data content of a specific ASSET_TYPE.

    All parameters must be specified by name.

    Parameters
    ----------
    identifier
        The identifier of the ASSET_TYPE asset.
    distribution_idx
        The index of a specific distribution from the distribution list (default is 0).
    version
        The version of the endpoint (default is None).

    Returns
    -------
    :
        The data content for the specified ASSET_TYPE.
    """
    url = url_to_get_content(asset_type, identifier, distribution_idx, version)
    res = requests.get(
        url,
        timeout=config.request_timeout_seconds,
    )
    distribution = res.content
    return distribution


def search(
    query: str,
    *,
    platforms: list[str] | None = None,
    offset: int = 0,
    limit: int = 10,
    search_field: (
        None | Literal["name", "issn", "description_html", "description_plain"]
    ) = None,
    get_all: bool = True,
    version: str | None = None,
    data_format: Literal["pandas", "json"] = "pandas",
    asset_type: str,
) -> pd.DataFrame | list[dict]:
    """Search metadata for ASSET_TYPE type using the Elasticsearch endpoint of the AIoD metadata catalogue.

    All parameters except `query` must be specified by name.

    Parameters
    ----------
    search
        The string to be matched against the search fields.
    platforms
        The platforms to filter the search results.
        If None, results from all platforms will be returned (default is None).
    offset
        The offset for pagination (default is 0).
    limit
        The maximum number of results to retrieve (default is 10).
    search_field
        The specific fields to search within. If None, the query will be matched against all fields (default is None).
    get_all
        If true, a request to the database is made to retrieve all data.
        If false, only the indexed information is returned. (default is True).
    version
        The version of the endpoint to use (default is None).
    data_format
        The desired format for the response (default is "pandas").
        For "json" formats, the returned type is a json decoded type, in this case a list of dict's.

    Returns
    -------
    :
        The retrieved metadata in the specified format.
    """
    url = url_to_search(
        asset_type,
        query,
        platforms,
        offset,
        limit,
        search_field,
        get_all,
        version,
    )
    res = requests.get(
        url,
        timeout=config.request_timeout_seconds,
    )
    resources = format_response(res.json()["resources"], data_format)
    return resources


async def get_assets_async(
    identifiers: list[str],
    *,
    asset_type: str,
    version: str | None = None,
    data_format: Literal["pandas", "json"] = "pandas",
) -> pd.DataFrame | list[dict]:
    """Asynchronously retrieve metadata for a list of ASSET_TYPE identifiers.

    All parameters except `identifiers` must be specified by name.

    Parameters
    ----------
    identifiers
        The list of identifiers of the ASSET_TYPE to retrieve.
    version
        The version of the endpoint (default is None).
    data_format
        The desired format for the response (default is "pandas").
        For "json" formats, the returned type is a json decoded type, in this case a list of dicts.

    Returns
    -------
    :
        The retrieved metadata for the specified ASSET_TYPE.
    """
    urls = [
        url_to_get_asset(asset_type, identifier, version) for identifier in identifiers
    ]
    response_data = await _fetch_resources(urls)
    resources = format_response(response_data, data_format)
    return resources


async def get_list_async(
    *,
    asset_type: str,
    offset: int = 0,
    limit: int = 100,
    batch_size: int = 10,
    version: str | None = None,
    data_format: Literal["pandas", "json"] = "pandas",
) -> pd.DataFrame | list[dict]:
    """Asynchronously retrieve a list of ASSET_TYPE from the catalogue in batches.

    All parameters must be specified by name.

    Parameters
    ----------
        offset: The offset for pagination (default is 0).
        limit: The maximum number of items to retrieve (default is 10).
        batch_size: The number of items in a a batch.
        version: The version of the endpoint (default is None).
        data_format: The desired format for the response (default is "pandas").
            For "json" formats, the returned type is a json decoded type, in this case a list of dicts.

    Returns
    -------
    :
        The retrieved metadata in the specified format.

    Raises
    ------
    ValueError
        Batch size must be larger than 0.
    """
    if batch_size <= 0:
        raise ValueError(
            "batch_size must be larger than 0, otherwise you can use the synchronous get_list function!"
        )

    offsets = range(offset, offset + limit, batch_size)
    last_batch_size = (limit % batch_size) or batch_size
    batch_sizes = [batch_size] * (len(offsets) - 1) + [last_batch_size]

    urls = [
        url_to_get_list(asset_type, offset, limit, version)
        for offset, limit in zip(offsets, batch_sizes)
    ]

    response_data = await _fetch_resources(urls)
    flattened_response_data = [
        response for batch in response_data for response in batch
    ]
    resources = format_response(flattened_response_data, data_format)
    return resources


async def _fetch_resources(urls) -> dict:
    async def _fetch_data(session, url) -> dict:
        async with session.get(url, timeout=config.request_timeout_seconds) as response:
            return await response.json()

    async with aiohttp.ClientSession() as session:
        tasks = [_fetch_data(session, url) for url in urls]
        response_data = await asyncio.gather(*tasks)
    return response_data


wrap_common_calls = partial(
    wrap_calls,
    calls=[
        get_list,
        counts,
        get_asset,
        post_asset,
        put_asset,
        patch_asset,
        delete_asset,
        get_asset_from_platform,
        get_content,
        get_assets_async,
        get_list_async,
    ],
)
wrap_search_call = partial(wrap_calls, calls=[search])



"""Tests which connect to the server.

Do not add new tests unless there is a very good reason.
These tests do not run with the default test configuration.

"""

import pytest

import aiod

DEFAULT_MARKER = "default"
LATEST_VERSION_MARKER = ""


@pytest.mark.server()
@pytest.mark.parametrize("version", [DEFAULT_MARKER, LATEST_VERSION_MARKER])
def test_get_dataset_list_from_default_version_server(version: str):
    if version == LATEST_VERSION_MARKER:
        aiod.config.version = LATEST_VERSION_MARKER
    datasets = aiod.datasets.get_list()
    assert len(datasets) == 10
    dataset = aiod.datasets.get_asset(datasets["identifier"].iloc[0])
    assert dataset["name"] == datasets["name"].iloc[0]



import pytest
import responses
from responses import matchers

import aiod.bookmarks


@responses.activate
def test_bookmark_create(valid_refresh_token):
    responses.post(
        "http://not.set/not_set/bookmarks?resource_identifier=valid_identifier",
        json={
            "resource_identifier": "valid_identifier",
            "created_at": "2025-09-23T13:01:14"
        },
        match=[matchers.header_matcher({"Authorization": "Bearer valid_access"})],
    )
    bookmark = aiod.bookmarks.register("valid_identifier")
    assert bookmark.identifier == "valid_identifier"
    assert bookmark.created.isoformat() == "2025-09-23T13:01:14+00:00"


@responses.activate
def test_bookmark_create_invalid_identifier(valid_refresh_token):
    responses.post(
        "http://not.set/not_set/bookmarks?resource_identifier=not_an_identifier",
        json={
            "detail": "Resource not_an_identifier does not exist.",
            "reference": "9745791405d945c084865ea2c3e4bc7b"
        },
        match=[matchers.header_matcher({"Authorization": "Bearer valid_access"})],
    )
    with pytest.raises(KeyError):
        aiod.bookmarks.register("not_an_identifier")


@responses.activate
def test_bookmark_delete(valid_refresh_token):
    responses.delete(
        "http://not.set/not_set/bookmarks?resource_identifier=valid_identifier",
        match=[matchers.header_matcher({"Authorization": "Bearer valid_access"})],
    )
    aiod.bookmarks.delete("valid_identifier")


@responses.activate
def test_bookmark_list(valid_refresh_token):
    responses.get(
        "http://not.set/not_set/bookmarks",
        json=[
          {
            "resource_identifier": "data_04BadJ7gCY3CZIR2Y1QDG8Z2",
            "created_at": "2025-09-23T13:01:14"
          },
          {
            "resource_identifier": "mdl_AbFo3gwAWsQ3HxLG90RC2GWH",
            "created_at": "2025-09-22T14:10:09"
          }
        ],
        match=[matchers.header_matcher({"Authorization": "Bearer valid_access"})],
    )
    bookmarks = aiod.bookmarks.get_list()
    assert len(bookmarks) == 2, "Each element should be included in the response"
    assert bookmarks[0].identifier.startswith("data"), "Each element should be parsed to a Bookmark"



import textwrap
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
import tomlkit
from typing import Any, Callable, TypeAlias
import logging

logger = logging.getLogger(__file__)
AttributeObserver: TypeAlias = Callable[[str, Any, Any], None]


@dataclass
class Config:
    """Configuration for which servers and versions to use in connections.

    Notes
    -----
    For most users, there is generally no reason to modify these values other
    than `version`.
    For developers, it may be useful to connect to a test server or local instance
    by changing the `api_server` and `auth_server` addresses.

    Attributes
    ----------
    version: str
        The version prefix to use.
    api_server: str
        The URL of the metadata catalogue REST API.
    auth_server: str
        The authentication server to connect to.
    realm: str
        The realm in which the client matching the `client_id` resides.
    client_id: str
        The client ID used for authentication.
    request_timeout_seconds: int
        If any request remains unresponsive for `request_timeout_seconds` seconds,
        it will automatically be aborted and raise a `requests.Timeout` error.

    """

    api_server: str = "https://api.aiod.eu/"
    version: str = "v2"
    auth_server: str = "https://auth.aiod.eu/aiod-auth/"
    realm: str = "aiod"
    client_id: str = "aiod-sdk"
    request_timeout_seconds: int = 10

    _observers: dict[str, set[AttributeObserver]] = field(
        default_factory=lambda: defaultdict(set),
        repr=False,
    )

    def subscribe(self, attribute: str, on_change: AttributeObserver) -> None:
        """Register a callback to be notified if the value of the `attribute` changes."""
        if on_change not in self._observers[attribute]:
            self._observers[attribute].add(on_change)

    def _store_to_file(self, attribute: str) -> None:
        if attribute not in dir(self):
            raise AttributeError("Cannot store ")
        user_config = tomlkit.loads(_user_config_file.read_text())
        if value := getattr(self, attribute):
            user_config[attribute] = value
        else:
            del user_config[attribute]

        _user_config_file.write_text(tomlkit.dumps(user_config))

    def __setattr__(self, key, value) -> None:
        old_value = getattr(self, key, None)
        super().__setattr__(key, value)

        if hasattr(self, "_observers"):
            for observer in self._observers[key]:
                observer(key, old_value, value)

    def _use_localhost(self, version: str = "latest"):
        """Set server URLs to localhost services. Convenience for developers."""
        self.version = version if version != "latest" else ""
        self.auth_server = "http://localhost/aiod-auth/"
        self.api_server = "http://localhost/"


def load_configuration(file: Path) -> Config:
    """Load a configuration from file.

    Parameters
    ----------
    file
        The TOML file that defines the configuration to load.

    Returns
    -------
    :
        The loaded configuration.

    """
    try:
        _user_config = tomlkit.loads(file.read_text())
    except tomlkit.exceptions.ParseError as e:
        _add_decode_error_note(file, e)
        raise

    global config
    for key, value in _user_config.items():
        setattr(config, key, value)
    return config


def _add_decode_error_note(file: Path, e: tomlkit.exceptions.ParseError) -> None:
    # Lines are 1-indexed and columns are 0-indexed
    e.add_note(
        textwrap.dedent(
            f"""
                Error reading configuration at {str(file)!r}: {e}
                File {str(file)!r}, line {e.line}:
                {file.read_text().splitlines()[e.line - 1]}
                {"^".rjust(e.col + 1)}
                """
        )
    )


config = Config()  # Modified through `load_configuration`
_user_config_file = Path("~/.aiod/config.toml").expanduser()
if _user_config_file.exists() and _user_config_file.is_file():
    load_configuration(_user_config_file)
    logger.info(f"Loaded configuration from {_user_config_file}: {config}.")
else:
    logger.info(
        f"No configuration file detected, using default configuration: {config}."
    )



from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin
from mkdocs.structure.files import Files
from mkdocs.structure.pages import Page

import logging

log = logging.getLogger(f"mkdocs.plugins.{__name__}")


class AssetTypeReplacer(BasePlugin):
    def on_page_content(
        self, html: str, page: Page, config: MkDocsConfig, files: Files
    ) -> str | None:
        if page.url.startswith("api/"):
            log.debug(f"Replacing ASSET_TYPE on {page.url}")
            html = html.replace("ASSET_TYPE", page.title.lower()).replace(
                '<span class="n">asset_type</span><span class="p">,</span>', ""
            )
        return html



import datetime
import time
import http.client
from http import HTTPStatus
import functools
from pathlib import Path

import requests
import tomlkit
from typing import Sequence, NamedTuple
from keycloak import KeycloakOpenID, KeycloakPostError, KeycloakConnectionError

from aiod.configuration import config
import logging

logger = logging.getLogger(__name__)

_token: "Token | None" = None
_user_token_file = Path("~/.aiod/token.toml").expanduser()


@functools.cache
def keycloak_openid() -> KeycloakOpenID:
    secret = (_token._client_secret or None) if _token else None
    return KeycloakOpenID(
        server_url=config.auth_server,
        client_id=config.client_id,
        realm_name=config.realm,
        client_secret_key=secret,
    )


def _on_keycloak_config_changed(_: str, __: str, ___: str) -> None:
    keycloak_openid.cache_clear()


config.subscribe("auth_server", on_change=_on_keycloak_config_changed)
config.subscribe("realm", on_change=_on_keycloak_config_changed)
config.subscribe("client_id", on_change=_on_keycloak_config_changed)


def _datetime_utc_in(*, seconds: int) -> datetime.datetime:
    span = datetime.timedelta(seconds=seconds)
    return datetime.datetime.now(datetime.UTC) + span


class Token:
    """Ensures active access tokens provided through one dedicated refresh token."""

    def __init__(
        self,
        *,
        client_secret: str | None = None,
        refresh_token: str | None = None,
        access_token: str | None = None,
        expires_in_seconds: int = -1,
    ):
        if (client_secret and refresh_token) or not (client_secret or refresh_token):
            raise ValueError(
                "Must set exactly one of `client_secret` or `refresh_token`."
            )
        if expires_in_seconds > 0 and access_token is None:
            raise ValueError(
                "If `expires_in_seconds` is set, `access_token` must be set to a valid access_token"
            )
        self._client_secret = client_secret
        self._refresh_token = refresh_token
        self._access_token = access_token or ""
        self._expiration_date = _datetime_utc_in(seconds=expires_in_seconds)

    @property
    def has_expired(self) -> bool:
        """Return whether the *access token* has expired based on local data."""
        return datetime.datetime.now(datetime.UTC) >= self._expiration_date

    @property
    def headers(self) -> dict[str, str]:
        """HTTP authorization header data for the token.

        Examples
        --------
        ```python
        import aiod
        token = aiod.get_token()
        requests.post(url, headers=token.headers, json=metadata)
        ```

        """
        if self.has_expired:
            self._refresh()
        return {"Authorization": f"Bearer {self._access_token}"}

    def __str__(self):
        return self._refresh_token

    def _refresh(self) -> None:
        """Use the `refresh token` or `client_secret` to request a new `access token`."""
        try:
            if self._refresh_token:
                token_info = keycloak_openid().refresh_token(self._refresh_token)
            if self._client_secret:
                token_info = keycloak_openid().token(grant_type="client_credentials")
        except KeycloakPostError:
            raise AuthenticationError(
                "Refresh token is not valid. Use `aiod.create_token` to get a new one."
            ) from None
        except KeycloakConnectionError as e:
            e.add_note(f"Could not connect {config.auth_server!r}, try again later.")
            raise

        self._access_token = token_info["access_token"]
        # self._refresh_token = token_info["refresh_token"]  # Only with auto-rotating
        # Because of the minuscule time difference between the server sending the
        # response and us processing it, the `expires_in` may not be used directly
        # when calculating expiration time.
        SAFETY_PERIOD_SECONDS = 2
        self._expiration_date = _datetime_utc_in(
            seconds=token_info["expires_in"] - SAFETY_PERIOD_SECONDS
        )
        logger.info(f"Renewed access token, it expires {self._expiration_date}.")

    def to_file(self, file: Path | None = None):
        doc = tomlkit.document()
        if self._refresh_token:
            doc.add("refresh_token", self._refresh_token)
        if self._client_secret:
            doc.add("client_secret", self._client_secret)
        if not self.has_expired:
            doc.add("access_token", self._access_token)
            doc.add("expiration_date", self._expiration_date.isoformat())

        file = file or _user_token_file
        file.parent.mkdir(parents=True, exist_ok=True)
        file.write_text(tomlkit.dumps(doc))

    @classmethod
    def from_file(cls, file: Path | None = None) -> "Token":
        file = file or _user_token_file
        doc = tomlkit.parse(file.read_text())
        kwargs: dict[str, str | int] = {
            "refresh_token": str(doc.get("refresh_token", "")),
            "client_secret": str(doc.get("client_secret", "")),
        }
        if "expiration_date" in doc:
            expiration_date = datetime.datetime.fromisoformat(doc["expiration_date"])
            expires_in = expiration_date - _datetime_utc_in(seconds=0)
            if expires_in.total_seconds() > 0:
                kwargs.update(
                    {
                        "access_token": str(doc["access_token"]),
                        "expires_in_seconds": expires_in.seconds,
                    }
                )
        return Token(**kwargs)  # type: ignore[arg-type]


def set_token(token: Token) -> None:
    """Set the token directly.

    Parameters
    ----------
    token
        Sets the token to be used for authenticated requests.

    Notes
    -----
    This function does not validate the provided token.
    If the token is invalid, subsequent authenticated requests may fail.
    """
    global _token
    _token = token
    keycloak_openid.cache_clear()


def get_token() -> Token:
    """Get the currently configured token that is used for authenticated requests.

    Returns
    -------
    :
    """
    if _token is None:
        msg = (
            "No token set. Please create a new token with `aiod.create_token()`,"
            " or set one with `aiod.set_token('...')`."
        )
        raise NotAuthenticatedError(msg)
    return _token


def _get_auth_headers(*, required: bool = True) -> dict:
    try:
        return get_token().headers
    except (AuthenticationError, NotAuthenticatedError):
        if required:
            raise
    return {}


def create_token(
    timeout_seconds: int = 300,
    *,
    write_to_file: bool = False,
    use_in_requests: bool = True,
) -> Token:
    """Get an API Key by prompting the user to log in through a browser.

    Notes
    -----
    This is a blocking function, and will poll the authentication server until
    authentication is completed or `timeout_seconds` have passed.

    Parameters
    ----------
    timeout_seconds
        The maximum time this function blocks waiting for the authentication workflow
        to complete. If `timeout_seconds` seconds have elapsed without successful
        authentication, this function raises an AuthenticationError.
        This must be set to a positive integer.
    write_to_file
        If set to true, the new api key (refresh token) will automatically be saved to
        the user configuration file (~/.aiod/config.toml).
    use_in_requests
        If set to true, the new token will automatically be used for future authenticated
        requests.

    Returns
    -------
    :
        The new token for use in authenticated requests.

    Raises
    ------
    AuthenticationError
        If authentication is unsuccessful in any way.

    """
    if timeout_seconds <= 0 or not isinstance(timeout_seconds, int):
        raise ValueError("`timeout_seconds` must be a positive integer.")
    kc = keycloak_openid()

    response = kc.device()
    device_code = response["device_code"]
    user_code = response["user_code"]
    verification_uri = response["verification_uri"]
    verification_uri_complete = response["verification_uri_complete"]

    # We print instead of log because this information *needs* to get to the user,
    # and relying on user logging configurations is too finnicky for that.
    print("Please authenticate using one of two methods:")  # noqa: T201
    print()  # noqa: T201
    print(f"  1. Navigate to {verification_uri_complete}")  # noqa: T201
    print(  # noqa: T201
        f"  2. Navigate to {verification_uri} and enter code {user_code}"
    )
    print()  # noqa: T201
    print(  # noqa: T201
        f"This workflow will automatically abort after {timeout_seconds} seconds."
    )

    poll_interval = response["interval"]
    start_time = time.time()

    token_endpoint = kc.well_known()["token_endpoint"]
    token_data = {
        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        "client_id": config.client_id,
        "device_code": device_code,
    }
    # We do not know when the user finishes their authentication, so we poll the server
    while time.time() - start_time < timeout_seconds:
        time.sleep(poll_interval)
        token_response = requests.post(
            token_endpoint,
            data=token_data,
            timeout=config.request_timeout_seconds,
        )
        token_response_data = token_response.json()

        response = (token_response.status_code, token_response_data.get("error"))
        match response:
            case (HTTPStatus.OK, _):
                access_token = token_response_data["access_token"]
                kc.decode_token(access_token, validate=True)
                token = Token(
                    refresh_token=token_response_data["refresh_token"],
                    access_token=access_token,
                    expires_in_seconds=token_response_data["expires_in"],
                )
                if write_to_file:
                    token.to_file(_user_token_file)
                if use_in_requests:
                    set_token(token)
                return token
            case (HTTPStatus.BAD_REQUEST, "authorization_pending"):
                continue
            case (HTTPStatus.BAD_REQUEST, "slow_down"):
                poll_interval *= 1.5
                continue
            case (HTTPStatus.BAD_REQUEST, "access_denied"):
                raise AuthenticationError("Access denied by Keycloak server.")
            case (HTTPStatus.BAD_REQUEST, "expired_token"):
                raise AuthenticationError("Device code has expired, please try again.")
            case (status, error):
                raise AuthenticationError(
                    f"Unexpected error, please contact the developers ({status}, {error})."
                )
    raise AuthenticationError(
        f"No successful authentication within {timeout_seconds=} seconds."
    )


def invalidate_token(
    token: str | Token | None = None, ignore_errors: bool = False
) -> None:
    """Invalidates the current (or provided) API key.

    Ends the current keycloak session, invalidating all keys issued.

    Parameters
    ----------
    token
        The token to invalidate.
        If str, it should be a refresh token.
        If None, it will default to the currently configured token.
    ignore_errors
        If true, do not raise an error if the logout attempt failed.
    """
    global _token
    token = token or _token
    try:
        keycloak_openid().logout(token)
    except (KeycloakPostError, KeycloakConnectionError) as e:
        if not ignore_errors:
            raise e
    finally:
        _token = None


class User(NamedTuple):
    name: str
    roles: Sequence[str]


def get_current_user() -> User:
    """Return name and roles of the user that is currently authenticated.

    Returns
    -------
    User
        The user information for the currently authenticated user.

    Raises
    ------
    NotAuthenticatedError
        When the user is not authenticated.
    """
    response = requests.get(
        f"{config.api_server}authorization_test",
        headers=get_token().headers,
        timeout=config.request_timeout_seconds,
    )

    content = response.json()
    if response.status_code == http.client.UNAUTHORIZED:
        raise NotAuthenticatedError(content)
    return User(
        name=content["name"],
        roles=tuple(content["roles"]),
    )


class AuthenticationError(Exception):
    """Raised when an authentication error occurred."""


class NotAuthenticatedError(Exception):
    """Raised when an endpoint that requires authentication is called without authentication."""


if _user_token_file.exists() and _user_token_file.is_file():
    try:
        _token = Token.from_file(_user_token_file)
    except Exception as e:
        e.add_note(f"Failed to load credentials from {str(_user_token_file)!r}")
        raise e



import pytest
import responses
from responses import matchers

from unittest.mock import Mock
import requests

import aiod
from aiod.configuration import config
from aiod.authentication.authentication import (
    keycloak_openid,
    AuthenticationError,
    create_token,
    get_token,
    set_token,
    Token,
    NotAuthenticatedError,
)


@pytest.fixture
def mocked_token() -> Mock:
    token = {"access_token": "fake_token", "refresh_token": "fake_refresh_token"}
    return Mock(return_value=token)


@pytest.fixture
def mocked_logout() -> Mock:
    return Mock()


def test_get_user_endpoint(mocked_token: Mock):
    keycloak_openid().token = mocked_token

    with responses.RequestsMock() as mocked_requests:
        mocked_requests.add(
            responses.GET,
            f"{config.api_server}authorization_test",
            json={"name": "user", "roles": ["a_role"]},
            status=200,
        )
        set_token(
            Token(refresh_token="fake_refresh", access_token="fake_token", expires_in_seconds=300)
        )
        user = aiod.get_current_user()

        header = mocked_requests.calls[0].request.headers["Authorization"]
        assert header == "Bearer fake_token", header
        assert user.name == "user", user
        assert user.roles == ("a_role",), user


def test_device_flow_success(monkeypatch):
    kc = keycloak_openid()
    kc.device = Mock(
        return_value={
            "device_code": "dev123",
            "user_code": "user123",
            "verification_uri": "http://verify",
            "verification_uri_complete": "http://verify?user=user123",
            "interval": 1,
        }
    )
    kc.well_known = Mock(
        return_value={"token_endpoint": "http://token", "jwks_uri": "http://jwks"}
    )

    success_response = Mock()
    success_response.status_code = 200
    success_response.json.return_value = {
        "access_token": "new_token",
        "refresh_token": "new_refresh",
        "expires_in": 300,
    }
    monkeypatch.setattr(requests, "post", lambda *a, **kw: success_response)
    kc.decode_token = Mock(return_value={"sub": "123"})

    create_token()
    assert get_token()._access_token == "new_token"
    assert get_token()._refresh_token == "new_refresh"


def test_device_flow_authorization_pending(monkeypatch):
    kc = keycloak_openid()
    kc.device = Mock(
        return_value={
            "device_code": "dev123",
            "user_code": "user123",
            "verification_uri": "http://verify",
            "verification_uri_complete": "http://verify?user=user123",
            "interval": 0,
        }
    )
    kc.well_known = Mock(
        return_value={"token_endpoint": "http://token", "jwks_uri": "http://jwks"}
    )

    # first response pending, second response success
    pending_response = Mock()
    pending_response.status_code = 400
    pending_response.json.return_value = {"error": "authorization_pending"}

    success_response = Mock()
    success_response.status_code = 200
    success_response.json.return_value = {
        "access_token": "token_ok",
        "refresh_token": "refresh_ok",
        "expires_in": 300,
    }

    calls = iter([pending_response, success_response])
    monkeypatch.setattr(requests, "post", lambda *a, **kw: next(calls))
    kc.decode_token = Mock(return_value={"sub": "123"})

    create_token()
    assert get_token()._access_token == "token_ok"
    assert get_token()._refresh_token == "refresh_ok"


def test_device_flow_failure(monkeypatch):
    kc = keycloak_openid()
    kc.device = Mock(
        return_value={
            "device_code": "dev123",
            "user_code": "user123",
            "verification_uri": "http://verify",
            "verification_uri_complete": "http://verify?user=user123",
            "interval": 1,
        }
    )
    kc.well_known = Mock(
        return_value={"token_endpoint": "http://token", "jwks_uri": "http://jwks"}
    )

    error_response = Mock()
    error_response.status_code = 400
    error_response.json.return_value = {"error": "expired_token"}
    monkeypatch.setattr(requests, "post", lambda *a, **kw: error_response)

    with pytest.raises(AuthenticationError):
        create_token()


def test_register_resource_no_token():
    with pytest.raises(NotAuthenticatedError):
        aiod.datasets.register(metadata=dict(name="Foo"))


@responses.activate
def test_register_resource_invalid_token(invalid_refresh_token):
    with pytest.raises(AuthenticationError) as e:
        aiod.datasets.register(metadata=dict(name="Foo"))
    assert "is not valid" in str(e.value)


@responses.activate
def test_register_resource_expired_token(expired_refresh_token):
    with pytest.raises(AuthenticationError) as e:
        aiod.datasets.register(metadata=dict(name="Foo"))
    assert "is not valid" in str(e.value)


@responses.activate
def test_register_resource_valid_token(valid_refresh_token):
    responses.post(
        "http://not.set/not_set/datasets",
        json={"identifier": "data_123412341234123412341234"},
        match=[matchers.header_matcher({"Authorization": "Bearer valid_access"})],
    )
    identifier = aiod.datasets.register(metadata=dict(name="Foo"))
    assert identifier == "data_123412341234123412341234"


def test_token_to_file_creates_parent_directory(tmp_path):
    token_file = tmp_path / ".aiod" / "token.toml"
    assert not token_file.parent.exists()

    token = Token(refresh_token="abc")
    token.to_file(token_file)
    assert token_file.parent.exists() and token_file.parent.is_dir()
    assert token_file.exists() and token_file.is_file()

    # Calling it multiple times should not result in an error,
    # even if the directory or file already exist.
    token.to_file(token_file)



"""Access any taxonomy defined by AI-on-Demand.

Some metadata fields in AI-on-Demand only accept a limited set of values.
These values are defined in *hierarchical taxonomies*.
The **taxonomy** defines terms with a unique name and a definition.
For example, in the taxonomy of 'Scientific Domains', one might define a
term 'Chemical Sciences' with the definition 'The scientific study of
the properties and behavior of matter.'.
The fact that the taxonomy is **hierarchical** means that each term may
have subterms defined which can further specify the entry. For example,
'Chemical Sciences' could be a subterm of 'Natural Sciences', with
'Biology' and 'Physics' being other subterms ("siblings").


Examples
--------
```python title="Minimal Example"
import aiod
aiod.taxonomies.news_categories()
```

```python title="Output"
[
  Term(taxonomy='news categories', term='Business', definition='', subterms=[
    Term(taxonomy='news categories', term='Funding', definition='', subterms=[]),
    ...,  # more subterms, which each may have subterms
  ]),
  ...,  # more terms
]
```

Notes
-----
Defined taxonomies differ by version of the REST API.
This module lists all taxonomies for all supported versions of the REST API.
If you access a taxonomy that is not available for your defined version, an
EndpointUndefinedError is raised.
"""

from __future__ import annotations

import dataclasses
import functools
import sys
from http import HTTPStatus
from typing import TypedDict

import requests

from aiod.configuration import config
from aiod.calls.urls import server_url
from aiod.calls.utils import EndpointUndefinedError

_mod = sys.modules[__name__]
_TAXONOMIES = [
    "licenses",
    "industrial_sectors",
    "news_categories",
    "number_of_employees",
    "publication_types",
    "research_areas",
    "scientific_domains",
    "turnovers",
]


@dataclasses.dataclass
class Term:
    """Describes a specific term in a hierarchical taxonomy.

    Attributes
    ----------
    taxonomy: str
        Name of the taxonomy from which the term originates, e.g., 'industrial sectors'.
    term: str
        A unique name for the term, e.g., 'Clinical Medicine'.
    definition: str
        A description further clarifying the meaning of the term, e.g.,
        'The branch of medicine that deals with the [...].'.
    subterms: list[Term]
        A list of subterms which provide a finer granularity.
        This list may be empty.
    """

    taxonomy: str
    term: str
    definition: str
    subterms: list[Term]

    def __eq__(self, other):
        return self.taxonomy == other.taxonomy and self.term == other.term


class _TermDict(TypedDict):
    term: str
    definition: str
    subterms: list[_TermDict]


def _parse_term(term: _TermDict, taxonomy: str) -> Term:
    subterms = [_parse_term(t, taxonomy) for t in term["subterms"]]
    return Term(
        taxonomy=taxonomy,
        term=term["term"],
        definition=term["definition"],
        subterms=subterms,
    )


def _get_taxonomy(name: str):
    # Since taxonomies rarely change, we cache the result in memory.
    @functools.cache
    def get_taxonomy() -> list[Term]:
        response = requests.get(
            f"{server_url()}{name}",
            timeout=config.request_timeout_seconds,
        )
        if response.status_code == HTTPStatus.NOT_FOUND:
            raise EndpointUndefinedError()
        if response.status_code != HTTPStatus.OK:
            raise RuntimeError(
                f"Unexpected response from ({response.status_code}, {response.json()})"
            )
        return [_parse_term(term, name.replace("_", " ")) for term in response.json()]

    get_taxonomy.__wrapped__.__doc__ = f"""
    Return the hierarchical taxonomy of {name.replace("_", " ")}.

    This function uses caching, and only the first call will result in a call to the server.
    The cache does not persist between Python sessions. You can clear the cache anytime by calling
    `aiod.taxonomies.{name}.cache_clear()`.

    Returns
    -------
    list[Term]
        A hierarchical taxonomy, each entry has a 'term', 'definition', and 'subterms'.
        The 'term' is a short unique name representing the entry.
        The 'definition' provides additional text to clarify the meaning for the term.
        'Subterms' provides terms which are part of this term, and may themselves have subterms as well.

    Raises
    ------
    EndpointUndefinedError
        If the taxonomy is not available for the configured version of the REST API.

    RuntimeError
        For unexpected server responses.
    """
    return get_taxonomy


for _taxonomy in _TAXONOMIES:
    setattr(_mod, _taxonomy, _get_taxonomy(_taxonomy))


__all__ = _TAXONOMIES + ["Term"]



import datetime as dt
from dataclasses import dataclass
from http import HTTPStatus

import requests

from aiod.configuration import config
from aiod.authentication import get_token
from aiod.calls.urls import server_url
from aiod.calls.utils import ServerError


@dataclass
class Bookmark:
    """Reference to an asset on AI-on-Demand.

    Attributes
    ----------
    identifier: str
        The identifier of the asset on AI-on-Demand, e.g., 'data_xyz...'
    created: datetime.datetime
        The datetime when the bookmark was originally created.
    """

    identifier: str
    created: dt.datetime


def _bookmarks_url() -> str:
    return server_url() + "bookmarks"


def register(identifier: str) -> Bookmark:
    """Bookmark the asset with `identifier` for the user.

    Parameters
    ----------
    identifier:
        The identifier of the asset on AI-on-Demand, e.g., 'data_xyz...'

    Returns
    -------
    :
        A `Bookmark` that denotes the time the bookmark was created.

    Raises
    ------
    KeyError
        If the identifier is not recognized by AI-on-Demand.
    ServerError
        If any other server-side error occurs.
    """
    res = requests.post(
        _bookmarks_url(),
        params=dict(resource_identifier=identifier),
        headers=get_token().headers,
        timeout=config.request_timeout_seconds,
    )
    if res.status_code == HTTPStatus.NOT_FOUND:
        raise KeyError(f"Could not find asset with identifier {identifier!r}.")
    if res.status_code != HTTPStatus.OK:
        raise ServerError(res)

    return Bookmark(
        identifier=res.json()["resource_identifier"],
        created=dt.datetime.fromisoformat(res.json()["created_at"]).replace(
            tzinfo=dt.UTC
        ),
    )


def delete(identifier: str):
    """Remove the bookmark for the asset with `identifier` for the user.

    This method does not raise an error if the identifier is not recognized by AI-on-Demand,
    or if the identifier is valid but the user has not bookmarked it.

    Parameters
    ----------
    identifier:
        The identifier of the asset on AI-on-Demand, e.g., 'data_xyz...'


    Raises
    ------
    ServerError
        If any other server-side error occurs.
    """
    res = requests.delete(
        _bookmarks_url(),
        params=dict(resource_identifier=identifier),
        headers=get_token().headers,
        timeout=config.request_timeout_seconds,
    )
    # The delete endpoint does not error if the bookmark does not exist
    if res.status_code != HTTPStatus.OK:
        raise ServerError(res)


def get_list() -> list[Bookmark]:
    """Return a list of the user's bookmarks.

    Returns
    -------
    :
        The list of bookmarks.
    """
    res = requests.get(
        _bookmarks_url(),
        headers=get_token().headers,
        timeout=config.request_timeout_seconds,
    )
    if res.status_code != HTTPStatus.OK:
        raise ServerError(res)
    return [
        Bookmark(
            identifier=bookmark["resource_identifier"],
            created=dt.datetime.fromisoformat(bookmark["created_at"]).replace(
                tzinfo=dt.UTC
            ),
        )
        for bookmark in res.json()
    ]
